"""Linear module.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""
from typing import Union

import torch
from torch import nn

from kindle.torch_utils import Activation


class Linear(nn.Module):
    """Linear module.

    Arguments: [channel, activation]
    """

    def __init__(self, in_channel: int, out_channel: int, activation: Union[str, None]):
        """

        Args:
            in_channels: input channels.
            out_channels: output channels.
            activation: activation name. If None is given, nn.Identity is applied
                which is no activation.
        """
        super().__init__()
        self.linear = nn.Linear(in_channel, out_channel)
        self.activation = Activation(activation)()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward."""
        return self.activation(self.linear(x))



"""Linear module generator.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""
from typing import List

import numpy as np

from kindle.generator.base_generator import GeneratorAbstract
from kindle.modules import Linear


class LinearGenerator(GeneratorAbstract):
    """Linear (fully connected) module generator for parsing."""

    def __init__(self, *args, **kwargs):
        """Initailize."""
        super().__init__(*args, **kwargs)

    @property
    def out_channel(self) -> int:
        """Get out channel size."""
        return self.args[0]

    @property
    def in_channel(self) -> int:
        """Get in channel size."""
        # error: Value of type "Optional[List[int]]" is not indexable
        return self.in_channels[self.from_idx]  # type: ignore

    def compute_out_shape(self, size: np.ndarray, repeat: int = 1) -> List[int]:
        """Compute output shape."""
        return [self.out_channel]

    def __call__(self, repeat: int = 1):
        # TODO: Apply repeat
        act = self.args[1] if len(self.args) > 1 else None

        return self._get_module(
            Linear(self.in_channel, self.out_channel, activation=act)
        )



"""Base Module Generator.

This module is responsible for GeneratorAbstract and ModuleGenerator.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""
from abc import ABC, abstractmethod
from typing import List, Tuple, Union

import numpy as np
from torch import nn

from kindle.torch_utils import make_divisible


class GeneratorAbstract(ABC):
    """Abstract Module Generator."""

    CHANNEL_DIVISOR: int = 8

    def __init__(
        self,
        *args,
        from_idx: Union[int, List[int]] = -1,
        in_channels: Tuple[int] = (0,),
        width_multiply: float = 1.0,
    ):
        """Initialize module generator.

        Args:
            *args: Module arguments
            from_idx: Module input index
            in_channels: Number of input channel
            width_multiply: Channel width multiply
        """
        self.args = tuple(args)
        self.from_idx = from_idx
        self.in_channels = in_channels
        self.width_multiply = width_multiply

    @property
    def name(self) -> str:
        """Module name."""
        return self.__class__.__name__.replace("Generator", "")

    def _get_module(self, module: Union[nn.Module, List[nn.Module]]) -> nn.Module:
        """Get module from __call__ function."""
        if isinstance(module, list):
            module = nn.Sequential(*module)

        # error: Incompatible types in assignment (expression has type "Union[Tensor, Module, int]",
        # variable has type "Union[Tensor, Module]")
        # error: List comprehension has incompatible type List[int];
        # expected List[Union[Tensor, Module]]
        module.n_params = sum([x.numel() for x in module.parameters()])  # type: ignore
        # error: Cannot assign to a method
        module.type = self.name  # type: ignore

        return module

    @classmethod
    def _get_divisible_channel(cls, n_channel: int) -> int:
        """Get divisible channel by default divisor.

        Args:
            n_channel: number of channel.

        Returns:
            Ex) given {n_channel} is 52 and {GeneratorAbstract.CHANNEL_DIVISOR} is 8.,
                return channel is 56 since ceil(52/8) = 7 and 7*8 = 56
        """
        return make_divisible(n_channel, divisor=cls.CHANNEL_DIVISOR)

    @property
    @abstractmethod
    def out_channel(self) -> int:
        """Out channel of the module."""

    @property
    @abstractmethod
    def in_channel(self) -> int:
        """In channel of the module."""

    @abstractmethod
    def compute_out_shape(self, size: np.ndarray, repeat: int = 1) -> List[int]:
        """Compute output shape when {size} is given.

        Args: input size to compute output shape.
        """

    @abstractmethod
    def __call__(self, repeat: int = 1):
        """Returns nn.Module component."""


class ModuleGenerator:
    """Module generator class."""

    def __init__(self, module_name: str):
        """Generate module based on the {module_name}

        Args:
            module_name: {module_name}Generator class must have been implemented.
        """
        self.module_name = module_name

    def __call__(self, *args, **kwargs):
        # replace getattr
        return getattr(
            __import__("kindle.generator", fromlist=[""]),
            f"{self.module_name}Generator",
        )(*args, **kwargs)



"""PyTorch Module Generator for parsing model yaml file.

- Author: Jongkuk Lim
- Contact: lim.jeikei@gmail.com
"""

from kindle.generator.base_generator import GeneratorAbstract, ModuleGenerator
from kindle.generator.bottleneck import BottleneckGenerator
from kindle.generator.concat import ConcatGenerator
from kindle.generator.conv import ConvGenerator
from kindle.generator.dwconv import DWConvGenerator
from kindle.generator.flatten import FlattenGenerator
from kindle.generator.linear import LinearGenerator
from kindle.generator.poolings import (AvgPoolGenerator,
                                       GlobalAvgPoolGenerator,
                                       MaxPoolGenerator)

__all__ = [
    "ModuleGenerator",
    "GeneratorAbstract",
    "BottleneckGenerator",
    "ConcatGenerator",
    "ConvGenerator",
    "DWConvGenerator",
    "FlattenGenerator",
    "LinearGenerator",
    "AvgPoolGenerator",
    "GlobalAvgPoolGenerator",
    "MaxPoolGenerator",
]



"""PyTorch Modules."""

from kindle.modules.bottleneck import Bottleneck
from kindle.modules.concat import Concat
from kindle.modules.conv import Conv
from kindle.modules.dwconv import DWConv
from kindle.modules.linear import Linear
from kindle.modules.poolings import GlobalAvgPool

__all__ = [
    "Bottleneck",
    "Concat",
    "Conv",
    "DWConv",
    "Linear",
    "GlobalAvgPool",
]
